

import argparse
from ctypes.wintypes import tagRECT
import datetime
import glob
import os
import re
import json

import t2g3lib

CarVelocity = 15	# [km/h] 平均速度がこれ以上は自動車

def main():
	parser = argparse.ArgumentParser(description="t2g3out")
	parser.add_argument('-o', '--output', help='file name to output')
	parser.add_argument('-c', '--only_car', action='store_true')
	parser.add_argument('-x', '--except_car', action='store_true')
	parser.add_argument('-t', '--find_title', help='find title in index.html')
	args = parser.parse_args()

	target_dirs = []

	# 対象ディレクトリの検索
	dirs = glob.glob('./*')
	for dir in dirs:
		if os.path.isdir(dir):
			file = os.path.join(dir, 'maps.files', 'vars.json')
			if os.path.exists(file):
				target_dirs.append(dir)

	# print(target_dirs)

	titles = {}
	if args.find_title:
		if os.path.exists(args.find_title):
			titles = read_index(args.find_title)

	total_cost = 0

	f = None
	if args.output:
		f = open(args.output, mode='w', encoding='utf-8')

	out = "Directory,Title,StartTime,EndTime,移動時間,実移動時間,移動距離,累積上り高度,累積下り高度,コース定数,消費カロリー,費用,実平均速度,平均速度,相対高度差,最大高度,最小高度,累積高度差,累積上り時間,累積下り時間,上り水平距離,下り水平距離,平均上り速度,平均下り速度,ピーク数,コースタイム,TimeRatio,TimeRatioPerc,ActualTimeRatio,ActualTimeRatioPerc"
	if args.output:
		f.write(out+'\n')
	else:
		print(out)
	for dir in target_dirs:
		# points, photos = loadInputData(os.path.join(dir, 'maps.files'))
		vars = loadVars(os.path.join(dir, 'maps.files'))
		if 'StartDateTime' in vars:
			# StartDateTimeがある場合のみ

			# 平均速度による自動車・登山判定
			iscar = False
			if 'car' in vars:
				iscar = vars['car']
			else:
				vel = vars['AvgSpeed']
				iscar = vel >= CarVelocity

			if args.only_car:
				if not iscar:
					continue
			if args.except_car:
				if iscar:
					continue

			keys = ['StartDateTime', 'EndDateTime', 'MovingTime', 'ActualMovingTime', 'MoveDistance', 'TotalUp', 'TotalDown', 'CourseConst', 'kcal', 'TotalCost', 'ActualAvgSpeed', 'AvgSpeed', 'RelAlt', 'MaxAlt', 'MinAlt', 'TotalAlt', 'TotalUpTime', 'TotalDownTime', 'TotalUpDistance', 'TotalDownDistance', 'TotalUpAvg', 'TotalDownAvg', 'NumOfPeaks', 'CourseTime', 'TimeRatio', 'TimeRatioPerc', 'ActualTimeRatio', 'ActualTimeRatioPerc']
			dir = dir.replace('.\\', '')	# .\の削除
			out = '<a href="'+dir+'">'+dir+'</a>'
			out += ','
			if dir in titles:
				out += titles[dir]
			for key in keys:
				if key in vars:
					s = str(vars[key])
					s = s.replace(',', '')
				else:
					s = ''
				out += ',' + str(s)
			if args.output:
				f.write(out+'\n')
			else:
				print(out)
		
		# parse comment.txt
		r, RPhotoNames, RTitle, RComment, MarkedFiles = t2g3lib.parseCommentFile(os.path.join(dir, 'comment.txt'))
		if r:
			name = 'cost'
			if name in RComment:
				lines = RComment[name].split("\n")
				# 合計金額を求める
				total = t2g3lib.parseCost(lines)
				# if total > 0:
				# 	print("{} cost={}".format(dir, total))
				total_cost += total

	if args.output:
		f.close()

	print("Cost={}円".format(comma(total_cost)))

# almost same function as defined in t2g3.py
def loadInputData(dir):
	with open(dir+'/points.json', 'r') as f:
	    points = json.load(f)
	with open(dir+'/photos.json', 'r') as f:
	    photos = json.load(f)

	# str -> DataTime conversion
	for pt in points:
		pt['time'] = str2datetime(pt['time'])
	for photo in photos:
		if 'time' in photo:
			photo['time'] = str2datetime(photo['time'])

	return points, photos

def loadVars(dir):
	file = dir+'/vars.json'
	if not os.path.exists(file):
		vars = {}
		return vars
	with open(file) as f:
		vars = json.load(f)
	return vars

def read_index(file):
	titles = {}
	for line in open(file, encoding='utf-8'):
		m = re.search(r'.+a href="(\d+)">(.+?)</a>', line, re.I)
		if m:
			dir = m.group(1)
			title = m.group(2)
			title = re.sub(r'<[^>]+>', '', title)	# tag除去
			title = re.sub(r'^[\d\/\-,]+ +', '', title)	# 先頭の日付削除
			if '周回' in title:
				print("dir={} title={}".format(dir, title))
			titles[dir] = title
	return titles

#
# Helpers
#

def str2datetime(str):
	try:
		return datetime.datetime.strptime(str, "%Y-%m-%dT%H:%M:%S%z")
	except:
		return datetime.datetime.strptime(str, "%Y-%m-%dT%H:%M:%S.%f%z")

def comma(val):
	return '{:,}'.format(val)

main()
