
import argparse
import os
import re
import datetime

def usage():
	print('''
parameter should be YYYYMMDD or YYYYMMDD-YYYYMMDD
''')

def main():
	MaxDays = 10

	parser = argparse.ArgumentParser(description="t2g3prep")
	parser.add_argument("date", nargs="?", help='YYYYMMDD or YYYYMMDD-YYYYMMDD')
	parser.add_argument("-m", '--max_days', help='maximum days')
	args = parser.parse_args()

	date = args.date
	if not date:
		print("Input date (YYYYMMDD)")
		now = datetime.datetime.now()
		date = "{:04}{:02}{:02}".format(now.year, now.month, now.day)
		print("Use " + date + " if only enter")
		r = input()
		if r != "":
			date = r

	print(date)

	if args.max_days:
		MaxDays = args.max_days

	m = re.match(r'^(\d\d\d\d)(\d\d)(\d\d)', date)
	if not m:
		usage()
		exit(1)

	start = datetime.datetime(year=int(m.group(1)), month=int(m.group(2)), day=int(m.group(3)))
	end = start
	m = re.match(r'.+-(\d\d\d\d)(\d\d)(\d\d)', date)
	if m:
		end = datetime.datetime(year=int(m.group(1)), month=int(m.group(2)), day=int(m.group(3)))

	dirs = []
	dt = start
	while dt <= end:
		dirs.append(dt.strftime('%Y%m%d'))
		dt += datetime.timedelta(days=1)

	if len(dirs) > MaxDays:
		print("Too long days")
		exit(10)

	for dir in dirs:
		print(dir)

		if os.path.exists(dir):
			print("Warning: Directory {} already exists.".format(dir))
			continue

		os.mkdir(dir)

	with open('Makefile.new', 'w') as fw:
		b_mode = False
		b_mode_follow = False
		for line in open('Makefile'):
			if b_mode:
				if not b_mode_follow:
					dir = dirs[0]
					line, n = re.subn(r' build_(\d+)', ' build_'+dir, line)	# 指定のdirectoryに変更
					if n == 0:
						print("Failed to find target b")
						exit(1)
					b_mode_follow = True
					fw.write(line)
					fw.write('\nbuild_{}:\n'.format(dir))
					for dir in dirs:
						fw.write('	$(MAKE) -DDIR={} buildnew\n'.format(dir))
					continue
				# 複数日対応の場合
				if line != '\n':
					# 非改行
					continue
				# 改行のみ
				b_mode = False
				b_mode_follow = False
	
			if line.startswith('b:'):
				b_mode = True

			# BUILDNEWSへの追加
			if line.startswith('BUILDNEWS ='):
				line = line.rstrip('\n')
				for dir in dirs:
					line += ' build_' + dir
				line += '\n'

			fw.write(line)

	if os.path.exists('Makefile.old'):
		os.unlink('Makefile.old')
	os.rename('Makefile', 'Makefile.old')
	os.rename('Makefile.new', 'Makefile')
	print("Makefile is changed.")


main()
